#
# MIT License
#
# Copyright (c) 2023 Killian Lucas
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#     ____                      ____      __                            __           
#    / __ \____  ___  ____     /  _/___  / /____  _________  ________  / /____  _____
#   / / / / __ \/ _ \/ __ \    / // __ \/ __/ _ \/ ___/ __ \/ ___/ _ \/ __/ _ \/ ___/
#  / /_/ / /_/ /  __/ / / /  _/ // / / / /_/  __/ /  / /_/ / /  /  __/ /_/  __/ /    
#  \____/ .___/\___/_/ /_/  /___/_/ /_/\__/\___/_/  / .___/_/   \___/\__/\___/_/     
#      /_/                                         /_/                               
#
# This file has been taken from the wonderful project "open-interpreter" by Killian Lucas
# https://github.com/KillianLucas/open-interpreter
#

import platform
from typing import List

from aiconsole.core.assets.materials.material import Material
from ..subprocess_code_interpreter import SubprocessCodeInterpreter
import os

class Shell(SubprocessCodeInterpreter):
    file_extension = "sh"
    proper_name = "Shell"

    def __init__(self):
        super().__init__()

        # Determine the start command based on the platform
        if platform.system() == 'Windows':
            self.start_cmd = 'cmd.exe'
        else:
            self.start_cmd = os.environ.get('SHELL', 'bash')

    def preprocess_code(self, code, materials: List[Material]):
        return preprocess_shell(code)
    
    def line_postprocessor(self, line):
        return line

    def detect_end_of_execution(self, line):
        return "## end_of_execution ##" in line
        

def preprocess_shell(code):
    """
    Add active line markers
    Wrap in a try except (trap in shell)
    Add end of execution marker
    """
    
    # Wrap in a trap for errors
    if platform.system() != 'Windows':
        code = wrap_in_trap(code)
    
    # Add end command (we'll be listening for this so we know when it ends)
    code += '\necho "## end_of_execution ##"'
    
    return code




def wrap_in_trap(code):
    """
    Wrap Bash code with a trap to catch errors and display them.
    """
    trap_code = """
trap 'echo "An error occurred on line $LINENO"; exit' ERR
set -E
"""
    return trap_code + code



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

import litellm
import tomlkit
import tomlkit.container
import tomlkit.exceptions
from aiconsole.api.websockets.outgoing_messages import DebugJSONWSMessage, SettingsWSMessage
from aiconsole.core.assets.asset import AssetStatus, AssetType
from aiconsole.core.project.paths import get_project_directory
from aiconsole.core.project.project import is_project_initialized
from aiconsole.utils.BatchingWatchDogHandler import BatchingWatchDogHandler
from aiconsole.utils.recursive_merge import recursive_merge
from appdirs import user_config_dir
from pydantic import BaseModel
from tomlkit import TOMLDocument
from watchdog.observers import Observer

_log = logging.getLogger(__name__)


class PartialSettingsData(BaseModel):
    code_autorun: Optional[bool] = None
    openai_api_key: Optional[str] = None
    disabled_materials: Optional[List[str]] = None
    disabled_agents: Optional[List[str]] = None
    enabled_materials: Optional[List[str]] = None
    enabled_agents: Optional[List[str]] = None
    forced_materials: Optional[List[str]] = None
    forced_agent: Optional[str] = None
    to_global: bool = False


class PartialSettingsAndToGlobal(PartialSettingsData):
    to_global: bool = False


class SettingsData(BaseModel):
    code_autorun: bool = False
    openai_api_key: Optional[str] = None
    disabled_materials: List[str] = []
    disabled_agents: List[str] = []
    enabled_materials: List[str] = []
    enabled_agents: List[str] = []
    forced_materials: List[str] = []
    forced_agent: str = ""


def _load_from_path(file_path: Path) -> Dict[str, Any]:
    with file_path.open() as file:
        document = tomlkit.loads(file.read())

    return dict(document)


def _set_openai_api_key_environment(settings: SettingsData) -> None:
    openai_api_key = settings.openai_api_key

    # This is so we make sure that it's overiden and does not env variables etc
    litellm.openai_key = openai_api_key or "invalid key"


class Settings:

    def __init__(self, project_path: Optional[Path] = None):
        self._suppress_notification_until = None
        self._settings = SettingsData()

        self._global_settings_file_path = Path(
            user_config_dir('aiconsole')) / "settings.toml"

        if project_path:
            self._project_settings_file_path = project_path / "settings.toml"
        else:
            self._project_settings_file_path = None

        self._observer = Observer()

        self._global_settings_file_path.parent.mkdir(
            parents=True, exist_ok=True)
        self._observer.schedule(
            BatchingWatchDogHandler(
                self.reload, self._global_settings_file_path.name),
            str(self._global_settings_file_path.parent),
            recursive=True)

        if self._project_settings_file_path:
            self._project_settings_file_path.parent.mkdir(
                parents=True, exist_ok=True)
            self._observer.schedule(
                BatchingWatchDogHandler(
                    self.reload, self._project_settings_file_path.name),
                str(self._project_settings_file_path.parent),
                recursive=True)

        self._observer.start()

    def model_dump(self) -> Dict[str, Any]:
        return self._settings.model_dump()

    def stop(self):
        self._observer.stop()

    async def reload(self, initial: bool = False):
        self._settings = await self.__load()
        await SettingsWSMessage(initial=initial or not (not self._suppress_notification_until or self._suppress_notification_until < datetime.datetime.now())).send_to_all()
        self._suppress_notification_until = None

    def get_asset_status(self, asset_type: AssetType, id: str) -> AssetStatus:
        s = self._settings

        if asset_type == AssetType.MATERIAL:
            if id in s.forced_materials:
                return AssetStatus.FORCED
            if id in s.disabled_materials:
                return AssetStatus.DISABLED
            return AssetStatus.ENABLED
        elif asset_type == AssetType.AGENT:
            if id == s.forced_agent:
                return AssetStatus.FORCED
            if id in s.disabled_agents:
                return AssetStatus.DISABLED
            return AssetStatus.ENABLED

        else:
            raise ValueError(f"Unknown asset type {asset_type}")

    def set_asset_status(self,
                         asset_type: AssetType,
                         id: str,
                         status: AssetStatus,
                         to_global: bool = False) -> None:

        if asset_type == AssetType.MATERIAL:
            partial_settings = {
                AssetStatus.DISABLED: PartialSettingsData(disabled_materials=[id]),
                AssetStatus.ENABLED: PartialSettingsData(enabled_materials=[id]),
                AssetStatus.FORCED: PartialSettingsData(
                    forced_materials=[id])
            }
        elif asset_type == AssetType.AGENT:
            partial_settings = {
                AssetStatus.DISABLED: PartialSettingsData(disabled_agents=[id]),
                AssetStatus.ENABLED: PartialSettingsData(enabled_agents=[id]),
                AssetStatus.FORCED: PartialSettingsData(forced_agent=id, enabled_agents=[
                                                        self._settings.forced_agent] if self._settings.forced_agent else None)
            }

        else:
            raise ValueError(f"Unknown asset type {asset_type}")

        self.save(partial_settings[status],
                  to_global=to_global)

    def get_code_autorun(self) -> bool:
        return self._settings.code_autorun

    def set_code_autorun(self,
                         code_autorun: bool,
                         to_global: bool = False) -> None:
        self._settings.code_autorun = code_autorun
        self.save(
            PartialSettingsData(code_autorun=self._settings.code_autorun),
            to_global=to_global)

    async def __load(self) -> SettingsData:
        settings = {}
        paths = [self._global_settings_file_path,
                 self._project_settings_file_path]

        for file_path in paths:
            if file_path and file_path.exists():
                settings = recursive_merge(
                    settings, _load_from_path(file_path))

        forced_agents = [
            agent for agent, status in settings.get('agents', {}).items()
            if status == AssetStatus.FORCED
        ]

        settings_data = SettingsData(
            code_autorun=settings.get('settings',
                                      {}).get('code_autorun', False),
            openai_api_key=settings.get(
                'settings', {}).get('openai_api_key', None),
            disabled_materials=[
                material
                for material, status in settings.get('materials', {}).items()
                if status == AssetStatus.DISABLED
            ],
            disabled_agents=[
                agent for agent, status in settings.get('agents', {}).items()
                if status == AssetStatus.DISABLED
            ],
            forced_materials=[
                material
                for material, status in settings.get('materials', {}).items()
                if status == AssetStatus.FORCED
            ],
            forced_agent=forced_agents[0] if forced_agents else "",)

        _set_openai_api_key_environment(settings_data)

        _log.info(f"Loaded settings")
        return settings_data

    @staticmethod
    def __get_tolmdocument_to_save(file_path: Path) -> TOMLDocument:
        if not file_path.exists():
            document = tomlkit.document()
            document['settings'] = tomlkit.table()
            document['materials'] = tomlkit.table()
            document['agents'] = tomlkit.table()
            return document

        with file_path.open() as file:
            document = tomlkit.loads(file.read())
            for section in ['settings', 'materials', 'agents']:
                if section not in dict(document):
                    document[section] = tomlkit.table()

        return document

    def save(self,
             settings_data: PartialSettingsData,
             to_global: bool = False) -> None:

        if to_global:
            file_path = self._global_settings_file_path
        else:
            if not self._project_settings_file_path:
                raise ValueError(
                    "Cannnot save to project settings file, because project is not initialized")

            file_path = self._project_settings_file_path

        document = self.__get_tolmdocument_to_save(file_path)
        doc_settings: tomlkit.table.Table = document['settings']
        doc_materials: tomlkit.table.Table = document['materials']
        doc_agents: tomlkit.table.Table = document['agents']

        if settings_data.code_autorun is not None:
            doc_settings['code_autorun'] = settings_data.code_autorun

        if settings_data.openai_api_key is not None:
            doc_settings['openai_api_key'] = settings_data.openai_api_key

        if settings_data.disabled_materials is not None:
            for material in settings_data.disabled_materials:
                doc_materials[material] = AssetStatus.DISABLED.value

        if settings_data.forced_materials is not None:
            for material in settings_data.forced_materials:
                doc_materials[material] = AssetStatus.FORCED.value

        if settings_data.enabled_materials is not None:
            for material in settings_data.enabled_materials:
                doc_materials[material] = AssetStatus.ENABLED.value

        if settings_data.disabled_agents is not None:
            for agent in settings_data.disabled_agents:
                doc_agents[agent] = AssetStatus.DISABLED.value

        if settings_data.forced_agent is not None:
            doc_agents[settings_data.forced_agent] = AssetStatus.FORCED.value

        if settings_data.enabled_agents is not None:
            for agent in settings_data.enabled_agents:
                doc_agents[agent] = AssetStatus.ENABLED.value

        self._suppress_notification_until = datetime.datetime.now() + \
            datetime.timedelta(seconds=30)

        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open('w') as file:
            file.write(document.as_string())


async def init():
    global _settings
    _settings = Settings(get_project_directory()
                         if is_project_initialized() else None)
    await _settings.reload()


def get_aiconsole_settings() -> Settings:
    return _settings


async def reload_settings(initial: bool = False):
    global _settings
    _settings.stop()
    _settings = Settings(get_project_directory()
                         if is_project_initialized() else None)
    await _settings.reload(initial)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
import asyncio
import traceback
from fastapi import APIRouter, Request
from fastapi.responses import StreamingResponse
import logging
from aiconsole.core.project import project
from aiconsole.core.code_running.run_code import get_code_interpreter
from aiconsole.core.code_running.types import CodeToRun
from aiconsole.utils.cancel_on_disconnect import cancelable_endpoint
from aiconsole.api.websockets.outgoing_messages import ErrorWSMessage

router = APIRouter()

_log = logging.getLogger(__name__)

@router.post("/chats/{chat_id}/run_code")
@cancelable_endpoint
async def run_code(request: Request, code_to_run: CodeToRun, chat_id: str) -> StreamingResponse:
    async def async_wrapper():
        try:
            _log.debug("Running code: %s", code_to_run.code)

            mats = [project.get_project_materials().get_asset(mid) for mid in code_to_run.materials_ids]

            # If the code starts with a !, that means a shell command
            if code_to_run.language == "python" and code_to_run.code.startswith("!"):
                code_to_run.language = "shell"
                code_to_run.code = code_to_run.code[1:]

            try:
                for chunk in get_code_interpreter(code_to_run.language).run(code_to_run.code, mats):
                    yield chunk
            except asyncio.CancelledError:
                get_code_interpreter(code_to_run.language).terminate()
                _log.info("Run cancelled")
            except Exception:
                await ErrorWSMessage(error=traceback.format_exc().strip()).send_to_chat(chat_id)
                yield traceback.format_exc().strip()
        except Exception as e:
            await ErrorWSMessage(error=str(e)).send_to_chat(chat_id)
            raise e

    return StreamingResponse(
        async_wrapper(), media_type="text/event-stream"
    )
    



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Optional, cast

from aiconsole.api.websockets.outgoing_messages import AnalysisUpdatedWSMessage
from aiconsole.consts import DIRECTOR_MIN_TOKENS, DIRECTOR_PREFERRED_TOKENS
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.analysis.create_plan_class import create_plan_class
from aiconsole.core.assets.asset import AssetLocation
from aiconsole.core.chat.types import Chat
from aiconsole.core.execution_modes.normal import execution_mode_normal
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.core.gpt.request import GPTRequest
from aiconsole.core.gpt.types import EnforcedFunctionCall, GPTMessage
from aiconsole.core.assets.materials.material import Material
from aiconsole.core.project import project
from aiconsole.utils.convert_messages import convert_messages
from pydantic import BaseModel

_log = logging.getLogger(__name__)


def pick_agent(arguments, chat: Chat, available_agents: list[Agent]) -> Agent:
    # Try support first
    default_agent = next(
        (agent for agent in available_agents if agent.id == "assistant"), None
    )

    # Pick any if not available
    if not default_agent:
        default_agent = available_agents[0]

    already_happened = arguments.already_happened
    is_users_turn = arguments.is_users_turn or already_happened

    if is_users_turn:
        picked_agent = Agent(
            id="user",
            name="User",
            usage="When a human user needs to respond",
            usage_examples=[],
            system="",
            defined_in=AssetLocation.AICONSOLE_CORE,
            gpt_mode=GPTMode.QUALITY,
        )
    else:
        picked_agent = next(
            (
                agent
                for agent in available_agents
                if agent.id == arguments.agent_id
            ),
            None,
        )

    _log.debug(f"Chosen agent: {picked_agent}")

    if not picked_agent:
        picked_agent = default_agent

    # If it turns out that the user must respond to him self, have the assistant drive the conversation
    if is_users_turn and chat.message_groups and chat.message_groups[-1].role == "user":
        picked_agent = default_agent

    return picked_agent


class AnalysisStepWithFunctionReturnValue(BaseModel):
    next_step: str
    picked_agent: Optional[Agent] = None
    relevant_materials: List[Material] = []


def _get_relevant_materials(relevant_material_ids: List[str]) -> List[Material]:
    # Maximum of 5 materials
    relevant_materials = [
        cast(Material, k)
        for k in project.get_project_materials().enabled_assets()
        if k.id in relevant_material_ids
    ][:5]

    relevant_materials += cast(list[Material], project.get_project_materials().forced_assets())

    return relevant_materials


async def gpt_analysis_function_step(
    chat: Chat,
    analysis_request_id: str,
    gpt_mode: GPTMode,
    initial_system_prompt: str,
    last_system_prompt: str,
    force_call: bool,
) -> AnalysisStepWithFunctionReturnValue:
    gpt_executor = GPTExecutor()

    # Pick from forced or enabled agents if no agent is forced
    forced_agents = project.get_project_agents().forced_assets()
    available_agents = forced_agents if forced_agents else project.get_project_agents().enabled_assets()
    available_agents = cast(list[Agent], available_agents)

    plan_class = create_plan_class(available_agents)

    request = GPTRequest(
        system_message=initial_system_prompt,
        gpt_mode=gpt_mode,
        messages=[*convert_messages(chat), GPTMessage(
            role="system",
            content=last_system_prompt
        )],
        functions=[plan_class.openai_schema],
        presence_penalty=2,
        min_tokens=DIRECTOR_MIN_TOKENS,
        preferred_tokens=DIRECTOR_PREFERRED_TOKENS,
    )

    if (force_call):
        request.function_call = EnforcedFunctionCall(name=plan_class.__name__)

    async for chunk in gpt_executor.execute(request):
        function_call = gpt_executor.partial_response.choices[0].message.function_call
        if function_call is not None:
            arguments = function_call.arguments
            if not isinstance(arguments, str):
                await AnalysisUpdatedWSMessage(
                    analysis_request_id=analysis_request_id,
                    agent_id=arguments.get("agent_id", None),
                    relevant_material_ids=arguments.get(
                        "relevant_material_ids", None),
                    next_step=arguments.get("next_step", None),
                    thinking_process=arguments.get("thinking_process", None),
                ).send_to_chat(chat.id)
        else:
            await AnalysisUpdatedWSMessage(
                analysis_request_id=analysis_request_id,
                agent_id=None,
                relevant_material_ids=None,
                next_step=None,
                thinking_process=gpt_executor.partial_response.choices[0].message.content,
            ).send_to_chat(chat.id)

    result = gpt_executor.response.choices[0].message

    if result.function_call is None:
        return AnalysisStepWithFunctionReturnValue(
            next_step=result.content or ""
        )

    arguments = result.function_call.arguments

    if isinstance(arguments, str):
        raise ValueError(
            f"Could not parse arguments from the text: {arguments}")

    arguments = plan_class(**arguments)

    picked_agent = pick_agent(arguments, chat, available_agents)

    relevant_materials = _get_relevant_materials(
        arguments.relevant_material_ids)

    return AnalysisStepWithFunctionReturnValue(
        next_step=arguments.next_step,
        picked_agent=picked_agent,
        relevant_materials=relevant_materials,
    )



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
import logging
from pathlib import Path
from fastapi import APIRouter
from fastapi.responses import FileResponse

from aiconsole.consts import AICONSOLE_PATH

router = APIRouter()

_log = logging.getLogger(__name__)


@router.get("/image")
async def image(path: str):
    abs_path = Path.cwd() / path

    if abs_path.exists():
        return FileResponse(str(abs_path))

    static_path = AICONSOLE_PATH / 'agents' / 'core' / 'default.jpg'
    return FileResponse(str(static_path))



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.assets.asset import AssetLocation, AssetStatus, AssetType
from aiconsole.core.gpt.consts import GPTMode
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from aiconsole.core.project import project
from aiconsole.core.settings.project_settings import get_aiconsole_settings

router = APIRouter()


@router.get("/")
async def agents_get():
    settings = get_aiconsole_settings()
    return JSONResponse(
        [
            *({
                **agent.model_dump(),
                "status": settings.get_asset_status(AssetType.AGENT, agent.id),
            } for agent in project.get_project_agents().all_assets())
        ]
    )




# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.agents.agent import Agent, AgentWithStatus
from aiconsole.core.assets.asset import AssetLocation, AssetStatus, AssetType
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.project import project
from aiconsole.core.settings.project_settings import get_aiconsole_settings
from fastapi import APIRouter, HTTPException
from fastapi.responses import JSONResponse
from pydantic import BaseModel

router = APIRouter()


class StatusChangePostBody(BaseModel):
    status: AssetStatus
    to_global: bool


@router.get("/{agent_id}")
async def agent_get(agent_id: str):
    if agent_id == "new":
        return JSONResponse(AgentWithStatus(
            id="",
            name="",
            usage="",
            usage_examples=[],
            status=AssetStatus.ENABLED,
            defined_in=AssetLocation.PROJECT_DIR,
            gpt_mode=GPTMode.QUALITY,
            system=""
        ).model_dump())
    else:
        try:
            settings = get_aiconsole_settings()
            agent = project.get_project_agents().get_asset(agent_id)
            return JSONResponse({
                **agent.model_dump(),
                "status": settings.get_asset_status(AssetType.AGENT, agent.id),
            })
        except KeyError:
            raise HTTPException(status_code=404, detail="Agent not found")
        


@router.patch("/{agent_id}")
async def agent_patch(agent_id: str, agent: Agent):
    try:
        await project.get_project_agents().save_asset(agent, new=False, old_asset_id=agent_id)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))

    return JSONResponse({"status": "ok"})


@router.post("/{agent_id}")
async def agent_post(agent_id: str, agent: Agent):
    try:
        await project.get_project_agents().save_asset(agent, new=True, old_asset_id=agent_id)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))

    return JSONResponse({"status": "ok"})


@router.post("/{agent_id}/status-change")
async def agent_status_change(
    agent_id: str,
    body: StatusChangePostBody
):
    """
    Change the status of a agent.

    Args:
        agent_id (str): The ID of the agent.
        body (StatusChangePostBody): POST body, only with "status"

    Returns:
        JSONResponse: JSON response indicating the result.
    """
    try:
        project.get_project_agents().get_asset(agent_id)
        get_aiconsole_settings().set_asset_status(
            AssetType.AGENT,
            id=agent_id, status=body.status, to_global=body.to_global
        )
        return JSONResponse({"status": "ok"})
    except KeyError:
        raise HTTPException(status_code=404, detail="Agent not found")
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))


@router.delete("/{agent_id}")
async def delete_agent(agent_id: str):
    try:
        project.get_project_agents().delete_asset(agent_id)
        return JSONResponse({"status": "ok"})
    except KeyError:
        raise HTTPException(status_code=404, detail="Agent not found")



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
import asyncio
import logging
from typing import AsyncGenerator
from aiconsole.core.assets.agents.agent import ExecutionModeContext
from aiconsole.core.execution_modes.get_agent_system_message import get_agent_system_message
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.utils.convert_messages import convert_messages
from aiconsole.core.gpt.create_full_prompt_with_materials import create_full_prompt_with_materials
from aiconsole.core.gpt.types import CLEAR_STR
from aiconsole.core.gpt.request import GPTRequest

_log = logging.getLogger(__name__)


async def execution_mode_normal(
    context: ExecutionModeContext,
) -> AsyncGenerator[str, None]:
    _log.debug("execution_mode_normal")

    gpt_executor = GPTExecutor()

    async for chunk in gpt_executor.execute(
        GPTRequest(
            messages=convert_messages(context.chat),
            gpt_mode=context.gpt_mode,
            system_message=create_full_prompt_with_materials(
                intro=get_agent_system_message(context.agent),
                materials=context.relevant_materials,
            ),
            min_tokens=250,
            preferred_tokens=2000,
        )
    ):
        if chunk == CLEAR_STR:
            yield chunk
            continue

        try:
            yield chunk["choices"][0]["delta"].get("content", "")
        except asyncio.CancelledError:
            _log.warning("Cancelled execution_mode_normal")
            break
